@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package JSON::Schema::Tiny; # git description: v0.003-10-g6eb5e0a
+package JSON::Schema::Tiny; # git description: v0.004-13-g4d862d6
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema, minimally
 # KEYWORDS: JSON Schema data validation structure specification tiny
 
-our $VERSION = '0.004';
+our $VERSION = '0.005';
 
 use 5.016;  # for the unicode_strings feature
 no if "$]" >= 5.031009, feature => 'indirect';
@@ -48,7 +48,7 @@ sub evaluate {
   shift
     if blessed($_[0]) and blessed($_[0])->isa(__PACKAGE__);
 
-  die 'insufficient arguments' if @_ < 2;
+  croak 'insufficient arguments' if @_ < 2;
   my ($data, $schema) = @_;
 
   my $state = {
@@ -90,7 +90,7 @@ sub evaluate {
 
 sub _eval {
   croak '_eval called in void context' if not defined wantarray;
-  die 'insufficient arguments' if @_ < 3;
+  croak 'insufficient arguments' if @_ < 3;
   my ($data, $schema, $state) = @_;
 
   # do not propagate upwards changes to depth, traversed paths,
@@ -118,7 +118,7 @@ sub _eval {
 
   foreach my $keyword (
     # CORE KEYWORDS
-    qw($schema $ref $defs),
+    qw($schema $ref $id $anchor $recursiveAnchor $vocabulary $comment $defs),
     # APPLICATOR KEYWORDS
     qw(allOf anyOf oneOf not if dependentSchemas
       items additionalItems contains
@@ -134,20 +134,24 @@ sub _eval {
     next if not exists $schema->{$keyword};
 
     $state->{keyword} = $keyword;
+    my $error_count = @{$state->{errors}};
 
     my $sub = __PACKAGE__->can('_eval_keyword_'.($keyword =~ s/^\$//r));
-    $valid = 0 if not $sub->($data, $schema, $state);
+    if (not $sub->($data, $schema, $state)) {
+      warn 'result is false but there are no errors (keyword: '.$keyword.')'
+        if $error_count == @{$state->{errors}};
+      $valid = 0;
+    }
 
-    warn 'result is false but there are no errors' if not $valid and not @{$state->{errors}};
     last if not $valid and $state->{short_circuit};
   }
 
   # UNSUPPORTED KEYWORDS
   foreach my $keyword (
     # CORE KEYWORDS
-    qw($id $anchor $recursiveAnchor $recursiveRef $vocabulary $dynamicAnchor $dynamicRef definitions),
+    '$recursiveRef',
     # APPLICATOR KEYWORDS
-    qw(dependencies prefixItems unevaluatedItems unevaluatedProperties),
+    qw(unevaluatedItems unevaluatedProperties),
   ) {
     next if not exists $schema->{$keyword};
     abort({ %$state, keyword => $keyword }, 'keyword not supported');
@@ -192,6 +196,65 @@ sub _eval_keyword_ref {
     });
 }
 
+sub _eval_keyword_id {
+  my ($data, $schema, $state) = @_;
+
+  return if not assert_keyword_type($state, $schema, 'string');
+
+  my $uri = Mojo::URL->new($schema->{'$id'});
+  abort($state, '$id value should not equal "%s"', $uri) if $uri eq '' or $uri eq '#';
+  abort($state, '$id value "%s" cannot have a non-empty fragment', $uri) if length $uri->fragment;
+  return 1;
+}
+
+sub _eval_keyword_anchor {
+  my ($data, $schema, $state) = @_;
+
+  return if not assert_keyword_type($state, $schema, 'string');
+
+  return 1 if ($schema->{'$anchor'}//'') =~ /^[A-Za-z][A-Za-z0-9_:.-]*$/;
+  abort($state, '%s value does not match required syntax', $state->{keyword});
+}
+
+sub _eval_keyword_recursiveAnchor {
+  my ($data, $schema, $state) = @_;
+
+  return if not assert_keyword_type($state, $schema, 'boolean');
+  return 1 if not $schema->{'$recursiveAnchor'};
+
+  # this is required because the location is used as the base URI for future resolution
+  # of $recursiveRef, and the fragment would be disregarded in the base
+  abort($state, '"$recursiveAnchor" keyword used without "$id"')
+    if not exists $schema->{'$id'};
+
+  return 1;
+}
+
+sub _eval_keyword_vocabulary {
+  my ($data, $schema, $state) = @_;
+
+  return if not assert_keyword_type($state, $schema, 'object');
+
+  foreach my $property (sort keys %{$schema->{'$vocabulary'}}) {
+    abort($state, '$vocabulary/%s value is not a boolean', $property)
+      if not is_type('boolean', $schema->{'$vocabulary'}{$property});
+
+    assert_uri($state, $schema, $property);
+  }
+
+  abort($state, '$vocabulary can only appear at the schema resource root')
+    if length($state->{schema_path});
+
+  abort($state, '$vocabulary can only appear at the document root')
+    if length($state->{traversed_schema_path}.$state->{schema_path});
+}
+
+sub _eval_keyword_comment {
+  my ($data, $schema, $state) = @_;
+  return if not assert_keyword_type($state, $schema, 'string');
+  return 1;
+}
+
 sub _eval_keyword_defs {
   my ($data, $schema, $state) = @_;
   return if not assert_keyword_type($state, $schema, 'object');
@@ -428,7 +491,7 @@ sub _eval_keyword_dependentRequired {
 
   return if not assert_keyword_type($state, $schema, 'object');
 
-  foreach my $property (keys %{$schema->{dependentRequired}}) {
+  foreach my $property (sort keys %{$schema->{dependentRequired}}) {
     E({ %$state, _schema_path_suffix => $property }, 'dependentRequired value is not an array'), next
       if not is_type('array', $schema->{dependentRequired}{$property});
 
@@ -444,7 +507,7 @@ sub _eval_keyword_dependentRequired {
   return 1 if not is_type('object', $data);
 
   my $valid = 1;
-  foreach my $property (keys %{$schema->{dependentRequired}}) {
+  foreach my $property (sort keys %{$schema->{dependentRequired}}) {
     next if not exists $data->{$property};
 
     if (my @missing = grep !exists($data->{$_}), @{$schema->{dependentRequired}{$property}}) {
@@ -878,7 +941,7 @@ sub is_equal {
   if ($types[0] eq 'object') {
     return 0 if keys %$x != keys %$y;
     return 0 if not is_equal([ sort keys %$x ], [ sort keys %$y ]);
-    foreach my $property (keys %$x) {
+    foreach my $property (sort keys %$x) {
       $state->{path} = jsonp($path, $property);
       return 0 if not is_equal($x->{$property}, $y->{$property}, $state);
     }
@@ -976,6 +1039,25 @@ sub assert_pattern {
   return 1;
 }
 
+sub assert_uri {
+  my ($state, $schema, $override) = @_;
+
+  my $string = $override // $schema->{$state->{keyword}};
+  my $uri = Mojo::URL->new($string);
+
+  abort($state, '"%s" is not a valid URI', $string)
+    # see also uri format sub
+    if fc($uri->to_unsafe_string) ne fc($string)
+      or $string =~ /[^[:ascii:]]/
+      or not $uri->is_abs
+      or $string =~ /#/
+        and $string !~ m{#$}                          # empty fragment
+        and $string !~ m{#[A-Za-z][A-Za-z0-9_:.-]*$}  # plain-name fragment
+        and $string !~ m{#/(?:[^~]|~[01])*$};         # json pointer fragment
+
+  return 1;
+}
+
 sub assert_non_negative_integer {
   my ($schema, $state) = @_;
 
@@ -1009,7 +1091,7 @@ JSON::Schema::Tiny - Validate data against a schema, minimally
 
 =head1 VERSION
 
-version 0.004
+version 0.005
 
 =head1 SYNOPSIS
 
@@ -1037,7 +1119,7 @@ specification. (See L</UNSUPPORTED JSON-SCHEMA FEATURES> below for exclusions.)
 =head1 FUNCTIONS
 
 =for Pod::Coverage is_type get_type is_equal is_elements_unique jsonp canonical_schema_uri E abort
-assert_keyword_type assert_pattern assert_non_negative_integer assert_array_schemas
+assert_keyword_type assert_pattern assert_uri assert_non_negative_integer assert_array_schemas
 new
 
 =head2 evaluate
@@ -1223,7 +1305,7 @@ L<Understanding JSON Schema|https://json-schema.org/understanding-json-schema>:
 
 Bugs may be submitted through L<https://github.com/karenetheridge/JSON-Schema-Tiny/issues>.
 
-I am also usually active on irc, as 'ether' at C<irc.perl.org> and C<irc.freenode.org>.
+I am also usually active on irc, as 'ether' at C<irc.perl.org> and C<irc.libera.chat>.
 
 =head1 AUTHOR
 
